
// Utilities

function identity (x) { return x; }
function equal (x,y) { return x == y; }
function constantly (x) { return function(){return x;} }
function propgetter (name) { return function(obj){ return obj[name]; } }

function afind (object, array, cmp, keyfn) {
  for (var i=0; i<array.length; i++) if (cmp(object, keyfn(array[i]))) return array[i];    
  return null;
}

function afind_from_end (object, array, cmp, keyfn) {
  for (var i=array.length-1; i>=0; i--) if (cmp(object, keyfn(array[i]))) return array[i];    
  return null;
}

// TODO: Learn enough javascript to make these generic.
function curry10 (fn,x) { return function() { return fn(x); } }
function curry21 (fn,x) { return function(y) { return fn(x,y); } }

function lcons (value, nextgenfn) {          // Lazy, memoizing list
  if (!nextgenfn) throw("wtf");
   return { value: value,
         toString: function() { return (this.value + (this.next()? (", " + this.next().toString()) : ""));},
             next: function() { var nextval = nextgenfn(); 
                               this.next = constantly(nextval);
                               return nextval; }}}

function stream_to_array (stream) {
  var a = new Array;
  while (stream) {
    a.push(stream.value);
    stream = stream.next();
  }
  return a;
}

// Printing

var debugging_parser = false;

function envelop (string, a, b) { return a + string + b; }
function parenthesize (string) { return envelop(string, "(", ")"); }

function subterm(parent,term) { 
  mywrite("subterm parent = " + parent + "  " + parent.name + "  (" + parent.precedence + " vs. " + term.precedence + ")");
  
  return (debugging_parser || (term.precedence < parent.precedence)) ? 
    parenthesize(term.toString()) :
    term.toString();
}

function postprinter (name, precedence) 
{ return function(operand) { return subterm({precedence: precedence}, operand.toString()) + name; } }

function argprinter (a, b) 
  { return function(op1,op2) { return op1 + envelop(op2, a, b); } }

// Language:
// Postfix unary operators have equal (and highest) precedence: (function application) [] !
// The prefix unary minus has the next highest precedence: -
// This is followed by the binary operators, grouped by precedence:
//   ^
//   * / 
//   + -
//   %
//   ,    (tuple append)
//   =

// Unary operators (FIXME)

var unary_factorial = { name: "!", descr: "Factorial", precedence: 10, printer: postprinter("!",10) };
var unary_aref = { descr: "ArrayRefrence", precedence: 10, printer: argprinter("[", "]") };
var unary_application = { descr: "Apply", precedence: 10,  printer: argprinter("(",")")};
var unary_minus = { descr: "Negate", precedence: 9, printer: function(x) { return "-" + subterm(this,x); } };

function left_associative  () { var state = null; return function(x) { var tmp=state; state=x; return tmp; }}
function right_associative () { var state = null; return function(x) { if (!state) state=x; return state; }}

function binop (name, descr, assoc) {
  return { name: name, descr: descr, assoc: assoc, 
      printer: function(x,y) { mywrite("binop printer " + this.name + ": precedence=" + this.precedence + "  " 
                                       + x.precedence + ":" +  x.toString() + "," + 
                                       y.precedence + ":" + y.toString()); 
                 //if(this.precedence == 4) throw("break");
                 return subterm(this,x) + name + subterm(this,y); } };
}

// Binary operators:
var binary_operators = [
 [ binop("=", "assignment",     right_associative) ],
 [ binop(",", "tuple append",   left_associative) ],
 [ binop("%", "modulus",        left_associative) ],
 [ binop("+", "addition",       left_associative),
   binop("-", "subtraction",    left_associative) ],
 [ binop("*", "multiplication", left_associative),
   binop("/", "division",       left_associative) ],
 [ binop("^", "exponentiation", left_associative) ]
];

// Initialize operator precedences according to table position
for (var i=0; i < binary_operators.length; i++) {
  for (var j=0; j < binary_operators[i].length; j++) {
    binary_operators[i][j].precedence = i;
  }
 }


// My parsing algorithm: Process unary operators and expressions nested in parens first,
// yielding an expression of the form "operand operator operand operator ... operand".
// Build the final AST from this by recursively partitioning the string at the operator
// of lowest precedence (and leftmost/rightmost position, according to associativity).
// Let's hope the user doesn't type in anything too long! =p

// Constructors for AST nodes:

function ast_unary (operator, operand) {
  return {
      operator: operator,
      precedence: operator.precedence,
      operand: operand,
      toString: curry10(operator.printer, operand),
      eval: curry21(operator.eval, operand) }; }

function ast_binary (operator, op1, op2) {
  return {
      operator: operator,
      precedence: operator.precedence,
      op1: op1,
      op2: op2,
      toString: function() { return operator.printer(op1,op2); },
      eval: function (env) { return operator.eval(op1, op2, env); } };}

function ast_number (value) { 
  return { 
      value: value, 
      precedence: 100, 
      toString: constantly(value.toString()), 
      eval: constantly(value) }; }

function ast_variable (name)  {
  throw("shouldn't be here!");
  return {
      name: name, 
      precedence: 100, 
      toString: constantly(name),  
      eval: function(env) { return env[this.name] || this; }}; }

function parse_error (reason) {
  mywrite("Parse error: " + reason);
  throw(reason);
}


// Lexical analysis

var whitespace_chars = " \r\n\t";
var numeric_chars = "0123456789.";
var op_chars = "+-*/%^=,()[]!";

function char_member (char, charset) { return !(-1 == charset.indexOf(char)); }
function char_whitespacep(char) { return char_member(char, whitespace_chars); }

function lex_token_body (stream, accum) {
  if (!stream) return null;
  else if (char_whitespacep(stream.value) || char_member(stream.value,op_chars)) return stream;
  else {
    accum.push(stream.value);
    return lex_token_body(stream.next(), accum);
  }
}

function lex_token_operator (stream, accum) {
  accum.push(stream.value);
  return stream.next();
}

function lex_token (stream, accum) {
  if (!stream) return null;
  else if (char_whitespacep(stream.value)) return lex_token(stream.next(), accum);
  else {
    if (numeric_chars.indexOf(stream.value) != -1) return lex_token_body(stream,accum); // Number
    else if (op_chars.indexOf(stream.value) != -1) return lex_token_operator(stream,accum); // Operator
    else return lex_token_body(stream,accum); // Identifier 
  }
}

function make_token_stream (charstream) {
  var accum = new Array;
  var cs_next = lex_token(charstream, accum);
  if (!accum.length) return null;
  else return lcons(accum.join(""), function(){return make_token_stream(cs_next)});
}


// Exprsession parser

function parse_level (tokens, start, end, optable, opindex) {
  var len = end-start;
  //  if (len == 1) 
  //  if (opindex == optable.length)
}

//function parse_expr (tokens, start, end) { return parse_level(tokens, start, end, binary_operators, 0); }
//function parse_toplevel (tokens) { return parse_expr(tokens, 0, tokens.length); }

// Parsing function return "pstates", which contain the value resulting from the parse, and the next unparsed cons in the token stream.

function pstate (value, rest) { return { value: value, rest: rest, toString: constantly("<Parser State>") }; }

// Cons an element onto the head of a (list-bearing) pstate
function pcons (elt, pst) { return pstate(lcons(elt, constantly(pst.value)), pst.rest); }
function pfstate (f, pst) { return pstate(f(pst.value), pst.rest); }

function parse_operand (string) {
  if (char_member(string[0], numeric_chars)) return ast_number(parseFloat(string)); // FIXME! Distinguish between floats and integers here.
  else return ast_variable(string);
}

function parse_unary_postfix (obj, list) {
  if (!list) return pstate(obj, null);
  else if (list.value == "!") return parse_unary_postfix(ast_unary(unary_factorial, obj), list.next());
  else if (list.value == "(") {
    var a = parse_flatten(list.next(), ")", false);
    return parse_unary_postfix(ast_binary(unary_application, obj, parse(a.value)), a.rest); // o_O
  } else return pstate(obj, list);
}

function parse_unary (list) {
  //mywrite("parse_unary: " + list);
  if (!list) parse_error("Operand expected");
  var x = list.value;
  if (x == "-") return pfstate(curry21(ast_unary, unary_minus), parse_unary(list.next()));
  else if (x == "(") {
    //mywrite("pu recursive parse");
    var a = parse_flatten(list.next(), ")", false);
    return parse_unary_postfix(parse(a.value), a.rest);
  } else return parse_unary_postfix(x, list.next());
}

function parse_flatten (list, nested, operatorMode) {
  //mywrite("parse_flatten (" + (operatorMode? "operator":"operand") +"): " + list);
  //if (list) mywrite("Next element is: " + list.next());
  if (nested && !list) parse_error("Expression ends prematurely: expected '" + nested + "'");
  else if (!list) return pstate(null,null);
  else if (operatorMode && nested && (list.value == nested)) return pstate(null, list.next());
  else if (operatorMode) return pcons(list.value, parse_flatten(list.next(), nested, false));
  else {
    var a = parse_unary(list);
    //mywrite("Appending tranformed element: " + a.value);
    return pcons(a.value, parse_flatten(a.rest, nested, true));
  }
}

function search_operators (list, end, comparefn, outputfn) {
  if (list && list.next() && list.next().next()) {
    assertfn(function(){return list != end;});
    assertfn(function(){return list.next() != end;});
    assertfn(function(){return list.next().next() != end;});
    if (comparefn(list.next().value)) outputfn(list.next());
    if (list.next().next().next() != end) return search_operators(list.next().next(), end, comparefn, outputfn);
    else return outputfn;
  } else parse_error("Wrong number of operands");
  throw("unreachable");
}

function findop (operators, string) { return afind(string, operators, equal, propgetter("name")); }

function parse_partitioning(list, end, precedence) {
  //mywrite("parse_partitioning p=" + precedence + ":  " + list.toString() + " upto " + end);
  if (list == end) parse_error("Missing operand");  //return null; 
  else if (precedence == binary_operators.length) {
    if (list.next() != null) parse_error("Unrecognized operator");
    else return list.value;
  } else if (list.next() == end) return list.value; //parse_operand(list.value);
  else {
    var precedence_group = binary_operators[precedence];
    var collector = precedence_group[0].assoc();
    var partition = search_operators(list, end, curry21(findop, precedence_group), collector)();

    if (partition == null) return parse_partitioning(list, end, precedence+1);
    else {
      var operator = assertnn(afind(partition.value, precedence_group, equal, propgetter("name")), "Operator lookup");
      return ast_binary(operator,
                        parse_partitioning(list, partition, 0),
                        parse_partitioning(partition.next(), end, 0));
    }
  }
}

function parse (list) {
  return parse_partitioning(list, null, 0);
}

function debug_dump_parse_ (p) {
  if (p.constructor == Array) {
    return p[0].descr + "(" + debug_dump_parse_(p[1]) + (p.length==3? ("," + debug_dump_parse_(p[2])): "") + ")";
  } else return p.toString();
}

function debug_dump_parse (p) {
  var s = debug_dump_parse_(p);

  var body = doc.getElementById("output");
  var paragraph = doc.createElement("P");
  var text = doc.createTextNode(s);
  paragraph.style.background="#FFFFFF";
  paragraph.style.font="12pt fixed";
  paragraph.appendChild(text);
  body.appendChild(paragraph);
}


function parse_and_eval (root) {
  var gen; // Generator for character stream
  gen = function(ednode) { 
    return ((!ednode)||(!edn_text(ednode))) ? 
      null : 
      lcons(edn_text(ednode), function(){return gen(edn_next(ednode))});
  }
  var tokens = make_token_stream(gen(root));
  mywrite("Token stream: " + tokens.toString());
  var parsed = parse(parse_flatten(tokens, null, false).value);
  //mywrite(parsed.toString());
  debug_dump_parse(parsed);
  //  parse_toplevel(stream_to_array(make_token_stream(gen(root))));
}
